# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import ClassVar
from typing import TypeVar

import pydantic
from libcanonical.bases import StateLogger


S = TypeVar('S', bound='Neuron')


class Neuron(pydantic.BaseModel, StateLogger):
    logger: ClassVar[logging.Logger] = logging.getLogger(__name__)
    active: bool = False
    coldkey: str
    consensus: float = 0.0
    dividends: float = 0.0
    emission: float = 0.0
    host: str = '0.0.0.0'
    hotkey: str
    incentive: float = 0.0
    port: int = 0
    rank: float = 0.0
    stake: float = 0.0
    trust: float = 0.0
    uid: int = -1
    updated: int = 0

    validator_permit: bool = pydantic.Field(
        default=False
    )

    vtrust: float = pydantic.Field(
        default=0.0
    )

    def is_miner(self) -> bool:
        raise NotImplementedError

    def is_validator(self) -> bool:
        return bool(self.validator_permit) and self.vtrust > 0.0

    def is_registered(self):
        return self.uid >= 0

    def diff(self, neuron: 'Neuron') -> set[str]:
        changed: set[str] = set()
        for attname in self.model_fields.keys():
            if getattr(self, attname) == getattr(neuron, attname):
                continue
            changed.add(attname)
        return changed

    def update(self, neuron: 'Neuron') -> tuple['Neuron', 'Neuron', tuple[str, ...]]:
        adapter: pydantic.TypeAdapter[Neuron] = pydantic.TypeAdapter(Neuron)
        changed = self.diff(neuron)
        old = adapter.validate_python(self.model_dump())
        for k in changed:
            new = getattr(neuron, k)
            setattr(self, k, new)
        return old, neuron, tuple(changed)

    def __hash__(self) -> int:
        return hash(self.hotkey)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ._coldkeyref import ColdkeyRef
from ._hotkey import Hotkey
from ._hotkeyset import HotkeySet
from ._hotkeyref import HotkeyRef
from ._hotkeypublickey import HotkeyPublicKey
from ._hotkeyprivatekey import HotkeyPrivateKey
from ._neuron import Neuron
from ._neuronlist import NeuronList
from ._persistedsynapse import PersistedSynapse
from ._synapseheader import SynapseHeader
from ._synapseenvelope import SynapseEnvelope


__all__: list[str] = [
    'ColdkeyRef',
    'Hotkey',
    'HotkeyPublicKey',
    'HotkeyPrivateKey',
    'HotkeyRef',
    'HotkeySet',
    'Neuron',
    'NeuronList',
    'PersistedSynapse',
    'SynapseHeader',
    'SynapseEnvelope',
]


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import threading
import os
from typing import cast
from typing import Any
from typing import Iterable
from typing import Literal
from typing import TYPE_CHECKING

import bittensor
import fastapi
from libcanonical.bases import StateLogger
from uvicorn import Server
from uvicorn import Config

from canonical.ext.bittensor.bases import BaseSoma
from canonical.ext.bittensor.models import Hotkey
from canonical.ext.bittensor.models import HotkeySet
from canonical.ext.bittensor.models import SynapseEnvelope
from ._apiroute import SynapseRouteHandler
from ._request import SynapseRequest
from ._router import SynapseRouter
if TYPE_CHECKING:
    from .._metagraph import Metagraph


AuthorizationFailureReason = Literal[
    'SIGNATURE_INVALID',
    'UNREGISTERED_NEURON',
    'INSUFFICIENT_STAKE',
    'NO_VALIDATOR_PERMIT',
]



class Axon(threading.Thread, StateLogger):
    __module__: str = 'canonical.ext.bittensor'
    allow_unregistered: bool = False
    force_validator_permit: bool = True
    min_stake: float = 4096.0
    soma: 'BaseSoma[Any]'

    def __init__(
        self,
        soma: 'BaseSoma[Any]',
        metagraph: 'Metagraph',
        hotkeys: list[Hotkey],
        host: str,
        port: int,
        synapse_types: Iterable[type[bittensor.Synapse]] | None = None,
        allow_unregistered: bool = False,
        force_validator_permit: bool = True,
        min_stake: float = 4096.0,
    ):
        super().__init__(
            daemon=True,
            target=self.main_event_loop
        )
        self.app = fastapi.FastAPI()
        self.allow_unregistered = allow_unregistered
        self.force_validator_permit  = force_validator_permit
        self.host = host
        self.port = port
        self.metagraph = metagraph
        self.hotkeys = HotkeySet(items={h.ss58_address: h for h in hotkeys})
        self.min_stake = min_stake
        self.running = threading.Event()
        self.soma = soma
        self.synapse_types = list(synapse_types or [])
        for synapse_class in self.synapse_types:
            self.attach(synapse_class)

    async def authorize(self, request: fastapi.Request, fatal: bool = True) -> bool:
        """Authorizes a :class:`~canonical.ext.bittensor.axon.SynapseRequest`."""
        request = cast(SynapseRequest[Any], request)
        conditions: list[AuthorizationFailureReason] = []

        # Verify the signature first.
        if not request.verify():
            conditions.append('SIGNATURE_INVALID')

        validator = self.metagraph.neuron(
            request.validator_hotkey,
            validator_permit=self.force_validator_permit
        )
        if not validator.is_registered() and not self.allow_unregistered:
            conditions.append('UNREGISTERED_NEURON')

        if (validator.stake < self.min_stake) and not self.allow_unregistered:
            # Unregistered neurons are assumed to never have stake in a
            # subnet.
            conditions.append('INSUFFICIENT_STAKE')

        if self.force_validator_permit and not validator.validator_permit:
            conditions.append('NO_VALIDATOR_PERMIT')

        for condition in conditions:
            match condition:
                case 'SIGNATURE_INVALID':
                    request.log(
                        'CRITICAL', "Signature did not validate (hotkey: %s)",
                        request.validator_hotkey
                    )
                case 'UNREGISTERED_NEURON':
                    request.log(
                        'CRITICAL', "Signer is not registered as a validator (hotkey: %s)",
                        request.validator_hotkey
                    )
                case 'INSUFFICIENT_STAKE':
                    request.log(
                        'CRITICAL', "Validator does not have sufficient stake (uid: %s)",
                        validator.uid
                    )
                case 'NO_VALIDATOR_PERMIT':
                    request.log(
                        'CRITICAL', "Validator does not have a validator permit (uid: %s)",
                        validator.uid
                    )

        if bool(conditions) and fatal:
            raise fastapi.HTTPException(
                status_code=403
            )

        return not bool(conditions)

    def attach(self, synapse_class: type[bittensor.Synapse]):
        """Exposes an endpoint to handle a synapse of the given class."""
        self.logger.info(
            "Accepting %s at http://%s:%s/%s",
            synapse_class.__name__,
            self.host,
            self.port,
            synapse_class.__name__
        )
        router: SynapseRouter[Any] = SynapseRouter(
            route_class=SynapseRouteHandler.build(synapse_class)
        )
        router.add_api_route(
            path=f'/{synapse_class.__name__}',
            endpoint=self.default_handler,
            methods=['POST'],
            response_model=synapse_class,
            response_description=(
                "The provided synapse updated with the results from the "
                "neuron."
            )
        )
        self.app.include_router(
            router=router,
            include_in_schema=True,
            dependencies=[
                fastapi.Depends(self.authorize)
            ]
        )

    async def default_handler(
        self,
        request: fastapi.Request
    ) -> fastapi.Response:
        return await self.submit(await cast(SynapseRequest[Any], request).envelope())

    def is_running(self):
        return self.running.is_set()

    def main_event_loop(self):
        self.config = Config(
            app=self.app,
            host=self.host,
            port=self.port,
            log_level='critical',
            access_log=False,
        )
        self.server = Server(config=self.config)
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        for synapse_class in self.synapse_types:
            self.attach(synapse_class)
        try:
            self.running.set()
            self.loop.run_until_complete(
                self.server.serve(sockets=[self.config.bind_socket()])
            )
        finally:
            self.running.clear()
            if self.config.uds and os.path.exists(self.config.uds):
                os.remove(self.config.uds)  # pragma: py-win32

    def stop(self):
        self.server.should_exit = True
        self.join()

    def submit(
        self,
        envelope: SynapseEnvelope[Any]
    ) -> asyncio.Future[fastapi.Response]:
        future = self.loop.create_future()
        self.soma.submit(envelope, future)
        return future

    def wait(self):
        return self.running.wait()


# Copyright (C) 2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import bittensor
import fastapi

from ._apiroute import SynapseRouteHandler


S = TypeVar('S', bound=bittensor.Synapse)


class SynapseRouter(fastapi.APIRouter, Generic[S]):
    """A :class:`fastapi.APIRouter` implementation that provided additional
    interfaces to register synapse routes.
    """
    __module__: str = 'tensorshield.ext.axon.routing'
    route_class: type[SynapseRouteHandler[S]]


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import threading
import os
from typing import Any
from typing import TypeVar
from typing import Union
from typing import TYPE_CHECKING

import bittensor
import fastapi
import uvicorn

from canonical.ext.bittensor.models import SynapseEnvelope
if TYPE_CHECKING:
    from canonical.ext.bittensor import Soma


U = TypeVar('U')


class AxonTransport(threading.Thread):
    """The :class:`~canonical.ext.bittensor.soma.AxonTransport`
    arranges communication between the :class:`~Soma` and the
    :class:`Axon`.
    """
    app: fastapi.FastAPI
    synapse_classes: tuple[type[bittensor.Synapse], ...]

    @staticmethod
    def union(classes: Any):
        return Union[tuple(classes)]

    def __init__(
        self,
        soma: 'Soma[Any]',
        synapse_classes: list[type[bittensor.Synapse]],
        host: str = '127.0.0.1',
        port: int = 8090
    ):
        super().__init__(
            daemon=True,
            target=self.main
        )
        self.soma = soma
        self.app = fastapi.FastAPI()
        self.host = host
        self.port = port
        self.running = threading.Event()
        self.synapse_classes = tuple(synapse_classes)
        self.config = uvicorn.Config(
            app=self.app,
            host=self.host,
            port=self.port,
            log_level='critical',
            access_log=False,
        )
        self.server = uvicorn.Server(config=self.config)

    def is_running(self):
        return self.running.is_set()

    def main(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        SynapseType = self.union(self.synapse_classes)

        async def handle(
            envelope: SynapseEnvelope[SynapseType] # type: ignore
        ) -> fastapi.Response:
            return await self.submit(envelope) # type: ignore

        self.app.add_api_route(
            endpoint=handle, # type: ignore
            path='/v1/synapses',
            methods=['POST']
        )

        try:
            self.running.set()
            self.loop.run_until_complete(
                self.server.serve(sockets=[self.config.bind_socket()])
            )
        finally:
            self.running.clear()
            if self.config.uds and os.path.exists(self.config.uds):
                os.remove(self.config.uds)  # pragma: py-win32

    def stop(self):
        self.server.should_exit = True
        self.join()

    def submit(
        self,
        envelope: SynapseEnvelope[Any]
    ) -> asyncio.Future[fastapi.Response]:
        future = self.loop.create_future()
        self.soma.submit(envelope, future)
        return future

    def wait(self):
        return self.running.wait()


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import TypedDict

from datasets import Sequence
from datasets import Value

from canonical.ext.datasets import Dataset



class PersistedSynapseDict(TypedDict):
    headers: dict[str, str]
    host: str
    kind: str
    method: str
    netuid: int
    path: str
    payload: str
    received: datetime.datetime


class PersistedSynapseDataset(Dataset[PersistedSynapseDict]):

    @classmethod
    def column_values(cls) -> dict[str, Sequence | Value]:
        return {
            'headers': Sequence(Sequence(Value('string'))),
            'host': Value('string'),
            'kind': Value('string'),
            'method': Value('string'),
            'netuid': Value('int32'),
            'path': Value('string'),
            'payload': Value('string'),
            'received': Value('timestamp[us]')
        }


# Copyright (C) 2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import bittensor
import fastapi


class SynapseResponse(fastapi.responses.JSONResponse):

    def __init__(
        self,
        synapse: bittensor.Synapse,
        status_code: int = 200
    ):
        if synapse.axon is not None:
            synapse.axon.status_code = status_code
            synapse.axon.status_message = "Success"
            if status_code >= 400:
                synapse.axon.status_message = "Error"
        super().__init__(
            status_code=status_code,
            content=synapse.model_dump(mode='json')
        )
        self.headers.update(synapse.to_headers()) # type: ignore
        self.headers['Content-Type'] = "application/json"


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ._const import *
from ._environment import BittensorEnvironment
from ._metagraph import Metagraph
from ._miner import BaseMiner
from ._validator import BaseValidator
from .axon import Axon
from .axon import SynapseRequest
from .models import *
from .soma import Soma
from .soma import Task


__all__: list[str] = [
    'Axon',
    'BaseMiner',
    'BaseValidator',
    'BittensorEnvironment',
    'Metagraph',
    'MinerNeuron',
    'NeuronList',
    'Soma',
    'Task',
    'SynapseRequest',
    'ValidatorNeuron',
]


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import time
import uuid
from typing import Any
from typing import Generic
from typing import NoReturn
from typing import TypeVar

import bittensor
import fastapi
import pydantic
from bittensor.core.settings import version_as_int as BITTENSOR_VERSION
from libcanonical.bases import StateLogger
from starlette.requests import empty_receive
from starlette.requests import empty_send
from starlette.types import Receive, Scope, Send

from canonical.ext.bittensor.models import SynapseEnvelope
from canonical.ext.bittensor.models import SynapseHeader


S = TypeVar('S', bound=bittensor.Synapse)


class SynapseRequest(fastapi.Request, StateLogger, Generic[S]):
    __module__: str = 'canonical.ext.bittensor.axon'
    _synapse: S | None
    logger = logging.getLogger(__name__)
    model: type[S]

    @property
    def validator_hotkey(self):
        return self.header.dendrite_hotkey

    def __init__(
        self,
        model: type[S],
        scope: Scope,
        receive: Receive = empty_receive,
        send: Send = empty_send
    ):
        super().__init__(scope, receive, send)
        self.model = model
        self.header = SynapseHeader.model_validate_headers(self.headers)
        self._synapse = None

    @property
    def synapse(self) -> S:
        if self._synapse is None:
            assert self.client is not None
            try:
                synapse = self.model.from_headers(dict(self.headers)) # type: ignore
            except pydantic.ValidationError:
                self.fail(422)
            else:
                if not synapse.axon:
                    self.fail(400, "Synapse did not specify a valid axon.")
                if not synapse.dendrite:
                    self.fail(400, "Synapse did not specify a valid dendrite.")
                synapse.axon.__dict__.update({
                    "version": str(BITTENSOR_VERSION),
                    "uuid": str(uuid.uuid4()),
                    "nonce": time.time_ns(),
                    "status_code": 100,
                })
                synapse.dendrite.__dict__.update({
                    "port": str(self.client.port),
                    "ip": str(self.client.host)
                })
                self._synapse = synapse # type: ignore
        assert self._synapse is not None
        return self._synapse

    def fail(self, status_code: int, message: str | None = "The service refused the request.") -> NoReturn:
        raise fastapi.HTTPException(status_code=status_code, detail=message)

    def get_logging_parameters(self) -> dict[str, Any]:
        return {
            'version': BITTENSOR_VERSION
        }

    def verify(self):
        if not self.synapse.dendrite:
            self.fail(400, "Synapse does not specify a dendrite.")
        if not self.synapse.dendrite.signature:
            self.fail(403, "Synapses must be signed by the sender.")
        message = str.join('.', [
            str(self.synapse.dendrite.nonce),
            str(self.synapse.dendrite.hotkey),
            str(self.header.axon_hotkey),
            str(self.synapse.dendrite.uuid),
            str(self.synapse.computed_body_hash)
        ])
        return self.header.dendrite_hotkey.verify(message, self.synapse.dendrite.signature)

    async def envelope(self) -> SynapseEnvelope[S]:
        assert self.client is not None
        return SynapseEnvelope(
            remote_host=self.client.host,
            remote_port=self.client.port,
            headers=self.header,
            synapse=self.model.model_validate_json(await self.body())
        )


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any
from typing import Generic
from typing import TypeVar

from canonical.ext.bittensor.models import SynapseEnvelope


S = TypeVar('S')


class BaseSoma(Generic[S]):
    """Specifies the base interface to a :class:`~canonical.ext.bittensor.Soma`
    implementation.
    """

    async def discover(self) -> None:
        raise NotImplementedError

    def submit(
        self,
        envelope: SynapseEnvelope[Any],
        future: asyncio.Future[Any]
    ) -> None:
        raise NotImplementedError


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
import pathlib
from typing import Literal

import pydantic

from ._hotkeyref import HotkeyRef
from ._hotkeyprivatekey import HotkeyPrivateKey
from ._hotkeypublickey import HotkeyPublicKey


class Hotkey(pydantic.RootModel[HotkeyRef|HotkeyPublicKey]):

    @property
    def public_bytes(self):
        if not isinstance(self.root, (HotkeyPublicKey, HotkeyPrivateKey)):
            raise TypeError("Hotkey is not loaded.")
        return self.root.public_key

    @property
    def hotkey(self):
        return self.root.hotkey

    @property
    def name(self):
        return self.root.name

    @property
    def ss58_address(self):
        if not isinstance(self.root, (HotkeyPublicKey, HotkeyPrivateKey)):
            raise TypeError("Hotkey is not loaded.")
        return self.root.ss58_address

    @property
    def qualname(self):
        return f'{self.root.name}/{self.root.hotkey}'

    def load(
        self,
        path: pathlib.Path,
        mode: Literal['public', 'private', 'phrase']
    ):
        p = path.joinpath(self.root.name, 'hotkeys', self.root.hotkey)
        with open(p, 'r') as f:
            data: dict[str, str] = {
                **json.loads(f.read()),
                'name': self.root.name,
                'hotkey': self.root.hotkey
            }
        match mode:
            case 'public': self.root = HotkeyPublicKey.model_validate(data)
            case 'private': self.root = HotkeyPrivateKey.model_validate(data)
            case 'phrase': raise NotImplementedError

        # TODO: Use stringwipe instead of garbage collect.
        del data

    def unload(self):
        self.root = HotkeyRef.model_validate(self.root.model_dump())

    def __hash__(self):
        return hash(self.root)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
import threading
import os
from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

import fastapi
import uvicorn

if TYPE_CHECKING:
    from canonical.ext.bittensor import Soma


U = TypeVar('U')


class Monitor(threading.Thread):
    app: fastapi.FastAPI

    def __init__(
        self,
        soma: 'Soma[Any]',
        host: str = '127.0.0.1',
        port: int = 8888,
    ):
        super().__init__(
            daemon=True,
            target=self.main_event_loop
        )
        self.app = fastapi.FastAPI()
        self.host = host
        self.soma = soma
        self.port = port

    async def live(self):
        return fastapi.Response(status_code=200)

    async def ready(self):
        status_code = 503
        if self.soma.step > 1 and not self.soma.must_exit:
            status_code = 200
        return fastapi.Response(status_code=status_code)

    async def startup(self):
        """Indicates a succesful startup of the application."""
        return fastapi.Response(status_code=200)

    def main_event_loop(self):
        self.loop = asyncio.new_event_loop()
        self.logger = logging.getLogger('canonical')
        self.config = uvicorn.Config(
            app=self.app,
            host=self.host,
            port=self.port,
            log_level='critical',
            access_log=False,
        )
        self.server = uvicorn.Server(config=self.config)

        self.logger.info(
            "Exposing liveness endpoint at http://%s:%s/live",
            self.host, self.port
        )
        self.app.add_api_route(
            endpoint=self.live,
            path='/live',
            include_in_schema=False,
            methods=['GET']
        )

        self.logger.info(
            "Exposing readyness endpoint at http://%s:%s/ready",
            self.host, self.port
        )
        self.app.add_api_route(
            endpoint=self.ready,
            path='/ready',
            include_in_schema=False,
            methods=['GET']
        )

        self.logger.info(
            "Exposing startup endpoint at http://%s:%s/startup",
            self.host, self.port
        )
        self.app.add_api_route(
            endpoint=self.startup,
            path='/startup',
            include_in_schema=False,
            methods=['GET']
        )

        try:
            asyncio.set_event_loop(self.loop)
            self.loop.run_until_complete(
                self.server.serve(sockets=[self.config.bind_socket()])
            )
        finally:
            if self.config.uds and os.path.exists(self.config.uds):
                os.remove(self.config.uds)  # pragma: py-win32

    def stop(self):
        self.server.should_exit = True
        self.join()


# Copyright (C) Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

import pydantic


class PersistedSynapse(pydantic.BaseModel):
    request_id: int = pydantic.Field(
        default=...
    )

    received: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    netuid: int = pydantic.Field(
        default=...
    )

    uid: int | None = pydantic.Field(
        default=None
    )

    kind: str = pydantic.Field(
        default=...
    )

    host: str = pydantic.Field(
        default=...
    )

    method: str = pydantic.Field(
        default=...
    )

    path: str = pydantic.Field(
        default=...
    )

    headers: dict[str, str] = pydantic.Field(
        default=...
    )

    payload: str | None = pydantic.Field(
        default=None
    )


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
import logging
import time
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import Union
from typing import TYPE_CHECKING

import bittensor
import fastapi

from canonical.ext.bittensor.axon import SynapseResponse
from canonical.ext.bittensor.models import SynapseEnvelope
if TYPE_CHECKING:
    from ._soma import SynapseTask
    from ._soma import SynapseHandlerType
    from ._taskrunner import TaskRunner


T = TypeVar('T', bound=bittensor.Synapse)


class Task(Generic[T]):
    handler: Union['SynapseHandlerType', 'SynapseTask']
    logger: logging.Logger = logging.getLogger(__name__)
    task_id: str
    subscribers: list[tuple[SynapseEnvelope[T], asyncio.Future[Any]]]

    def __init__(
        self,
        runner: 'TaskRunner',
        handler: 'SynapseHandlerType',
        task_id: str
    ):
        self.task_id = task_id
        self.created = time.time()
        self.handler = handler
        self.subscribers = []
        self.running = False
        self.runner = runner
        self.processed = 0
        self._lock = asyncio.Lock()

    def is_initial(self) -> bool:
        return self.processed == 0

    def is_runnable(self):
        return any([
            bool(self.subscribers) and not self.running,
        ])

    def lock(self):
        return self._lock

    def subscribe(
        self,
        envelope: SynapseEnvelope[T],
        future: asyncio.Future[SynapseResponse]
    ):
        self.logger.debug(
            "Creating subscription for %s (task-id: %s)",
            type(envelope.synapse).__name__,
            self.task_id[:6]
        )
        self.subscribers.append((envelope, future))

    async def age(self, min_age: int):
        """Block until the task has reached the given age, in seconds."""
        td = time.time() - self.created
        await asyncio.sleep(max(0, min_age - td))

    async def run(self) -> None:
        """Runs all subscribing synapses sequentially."""
        async with self.lock():
            self.logger.debug(
                "Running handler on all synapses (task-id: %s)",
                self.task_id[:6]
            )
            try:
                self.running = True
                await self._run()
            except Exception as e:
                self.logger.exception(
                    "Caught fatal %s while running handler: %s",
                    type(e).__name__,
                    repr(e)
                )
            finally:
                self.logger.debug(
                    "Finished running (task-id: %s, total: %s)",
                    self.task_id[:6],
                    self.processed
                )
                self.running = False

    async def _run(self):
        if inspect.isasyncgenfunction(self.handler):
            self.handler = await self.runner.run_handler(task=self, handler=self.handler)
            assert inspect.isasyncgen(self.handler)
            await self.handler.asend(None)
        while True:
            try:
                envelope, future = self.subscribers.pop()
            except IndexError:
                break
            try:
                if inspect.isasyncgen(self.handler):
                    result = await self.handler.asend(envelope.synapse)
                else:
                    result = await self.runner.run_handler(
                        self,
                        self.handler, # type: ignore
                        envelope=envelope
                    )
            except StopAsyncIteration:
                # Assume here that a fatal exception occurred earlier within the loop.
                result = SynapseResponse(
                    synapse=envelope.synapse,
                    status_code=500
                )
            except Exception as e:
                self.logger.exception(
                    'Caught fatal %s while running handler: %s',
                    type(e).__name__,
                    repr(e)
                )
                result = SynapseResponse(
                    synapse=envelope.synapse,
                    status_code=500
                )
            self.processed += 1
            if not isinstance(result, (bittensor.Synapse, fastapi.Response)):
                raise TypeError(
                    f"Handler for {type(envelope.synapse).__name__} did not return "
                    f"a bittensor.Synapse or fastapi.Response object: {repr(result)}."
                )
            if isinstance(result, bittensor.Synapse):
                result = SynapseResponse(synapse=result)
            self.runner.loop.call_soon_threadsafe(future.set_result, result)
        assert not self.subscribers

    def __await__(self):
        return self.run().__await__()


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import bittensor
import pydantic

from ._synapseheader import SynapseHeader


S = TypeVar('S', bound=bittensor.Synapse)


class SynapseEnvelope(pydantic.BaseModel, Generic[S]):
    remote_host: str
    remote_port: int | None = None
    headers: SynapseHeader
    synapse: S

    @classmethod
    def annotate(cls, synapse_class: type[S]) -> type['SynapseEnvelope[S]']:
        return SynapseEnvelope[synapse_class]